/* Copyrights 2002 Luis Figueiredo (stdio@netc.pt) All rights reserved. 
 *
 * See the LICENSE file
 *
 * The origin of this software must not be misrepresented, either by
 * explicit claim or by omission.  Since few users ever read sources,
 * credits must appear in the documentation.
 * 
 * date: Sat Mar 30 14:44:42 GMT 2002
 * 
 * -- client handler functions
 *
 */


#include "client.h"


extern char *_libwebserver_version; // Defined in server.c


struct web_client *current_web_client;

/*********************************************************************************************************/
/*
 * initializate (allocate) client list
 */
struct web_client *init_client_list() {
	struct web_client *ret;
	IFDEBUG(fprintf(stderr,"client.c: (allocating head client list): ");fflush(stderr);); // REMOVE
	ret=mymalloc(sizeof(struct web_client));
	if(ret==NULL) {
		fprintf(stderr,"client.c: Not enough mem to allocate client list\n");
		return NULL;
	};
#ifdef HAVE_OPENSSL
	ret->ssl=NULL; // ssl handler for this client
#endif
	ret->next=NULL;
	return ret;
}


/*********************************************************************************************************/
/*
 * Add a client node to client list
 */
int add_client(struct web_client *list, struct web_client *node) {
	struct web_client *temp=list;
#ifdef WIN32
	int t=IOC_INOUT;
#endif
	while(temp->next!=NULL)temp=temp->next; // run to last client
	temp->next=node;
	temp->next->rbuf=NULL;
	temp->next->rbufsize=temp->next->wrotesize=0;
	temp->next->outstream=init_outstream_list();
	if(temp->next->outstream==NULL) {
		fprintf(stderr,"client.c: Unable to create outstream list\n");
		return 0;
	};
	temp->next->newdata_try=0;
#ifdef WIN32
	ioctlsocket(temp->next->socket,FIONBIO,&t);  //non blocking sockets for win32
	
#else
	fcntl(temp->next->socket,F_SETFL,O_NONBLOCK);
#endif
	temp->next->next=NULL;
	temp->next->HTTPdirective=NULL;
	temp->next->stat=1; // Add a connected client
	IFDEBUG(fprintf(stderr,"client.c: Client added\n"));
	return 1;
}
/*********************************************************************************************************/

#ifdef WIN32
#define SHUT_RDWR SD_BOTH
#endif

/*********************************************************************************************************/
/*
 * Delete client node 
 */
void delete_client(struct web_client *node) {
	int rt;
	rt=shutdown(node->socket,SHUT_RDWR);
	IFDEBUG(fprintf(stderr,"(SEE) shutdown is %d\n",rt));
#ifdef WIN32
	rt=closesocket(node->socket); 
#else
	rt=close(node->socket); 
#endif
	
	IFDEBUG(fprintf(stderr,"close is %d\n",rt));
	delete_outstream_list(node->outstream);
#ifdef HAVE_OPENSSL
	SSL_free (node->ssl);
#endif
	IFDEBUG(fprintf(stderr,"client.c: (freeing client->rbuf): ");fflush(stderr);); // REMOVE
	myfree(node->rbuf); // free's
	IFDEBUG(fprintf(stderr,"client.c: (freeing client node): ");fflush(stderr);); // REMOVE
	myfree(node);       // free's
	IFDEBUG(fprintf(stderr,"client.c: Client delected\n"));
}


/*********************************************************************************************************/
/*
 * Delete next client node 
 */
void delete_next_client(struct web_client *node) {
	struct web_client *temp=node->next;
	node->next=node->next->next;
	delete_client(temp);
}

/*********************************************************************************************************/
/* 
 * Read what client have to say
 */
void read_client(struct web_client *node) {
	int tmp,tmp1;
	char *tmp2;
	char readtemp[READMAX];
	
	//if(tmp=myread(node->socket,readtemp,READMAX)>0) {
	if(newdata(node->socket)) { //Change
#ifdef HAVE_OPENSSL
		if(node->ssl!=NULL) {	
			tmp=SSL_read(node->ssl,readtemp,READMAX);		
		} else {
			tmp=myread(node->socket,readtemp,READMAX);
		};

#else
		tmp=myread(node->socket,readtemp,READMAX);
#endif
		if(tmp<1) {
			// non_block case....
			// #define EAGAIN 11
			if(errno!=11) node->stat=5;	//interesting
			return;	
		};
		tmp1=node->rbufsize;
		node->rbufsize+=tmp;
		IFDEBUG(if(node->rbuf==NULL)fprintf(stderr,"client.c: (allocating client->rbuf): ");fflush(stderr);); // REMOVE
		tmp2=myrealloc(node->rbuf,node->rbufsize+1);
		IFDEBUG(fprintf(stderr,"client.c: tmp is %d tmp2 is %p\n",tmp,tmp2));
		if(tmp2==NULL) {
			node->stat=5;
			return;
		}else {
			node->rbuf=tmp2;
		};
		memcpy(node->rbuf+tmp1,readtemp,tmp);
		node->rbuf[node->rbufsize]='\0';
		node->newdata_try=0;
	}else {
		IFDEBUG(fprintf(stderr,"client.c: try++\n"));
		node->newdata_try++;
		if((node->rbufsize>0 && (strstr((node->rbuf+node->rbufsize)-4,"\r\n\r\n") || strstr((node->rbuf+node->rbufsize)-4,"--\r\n"))) || node->newdata_try>100) {  // (NEW) if pass's 100 times without giving data (200 ms i guess, since select (uses 2ms)) then over
			IFDEBUG(fprintf(stderr,"client.c: Client all read (%d bytes)\n",node->rbufsize));		
			IFDEBUG(fprintf(stderr,"client.c: Read (%p):",node->rbuf));
			node->stat=2; // Next state
			if(node->rbufsize<1) {
				node->stat=5;
				return;
			}else {
				IFDEBUG(fwrite(node->rbuf,node->rbufsize,1,stderr));
			};
			IFDEBUG(fprintf(stderr,"\n"));
			//write(2,node->rbuf,node->rbufsize); //REMOVE DEBUG
			//current_web_server->stat.brecv+=node->rbufsize; // stats
		}
	};
}
/*********************************************************************************************************/
/*
 * Process headers w/ get handlers
 */
void process_client(struct web_client *node,struct gethandler *list) {
	struct gethandler *gettemp=list;
	int tmp=0;
	char *tmp1=web_client_getreq();
	char matchbuf[MATCHMAX];	
	FILE tfile; 
	FILE *nfile;
	char *fname;
	while(gettemp->next!=NULL && tmp==0) {
		gettemp=gettemp->next;
		snprintf(matchbuf,MATCHMAX,"%s",gettemp->str);
		if(!tmp1) {
			web_client_writef(node,"HTTP/1.1 400 Invalid request\r\n");
			web_client_writef(node,"Date: %s\r\n",mydate(time(NULL),"%a, %d %b %Y %H:%M:%S %Z")); // Date header
			web_client_writef(node,"Server: %s\r\n",_libwebserver_version);
			web_client_writef(node,"Content-type: text/html\r\n\r\n<HTML><title>Invalid request</title><body bgcolor=FFFFFF><font size=6>400 Invalid  request</font><BR><BR><HR><small><i>%s</i></small></body></html>\n\r",_libwebserver_version);                                    
			web_log("%s - - [%s] ^ Invalid request\n",inet_ntoa(node->sa.sin_addr),mydate(time(NULL),"%d/%b/%Y:%H:%M:%S %z"));
			node->stat=5;
			return;
		};
		if(strlen(tmp1)>MAXURLSIZE) {
			web_client_writef(node,"HTTP/1.1 414 URL to large\r\n");
			web_client_writef(node,"Date: %s\r\n",mydate(time(NULL),"%a, %d %b %Y %H:%M:%S %Z")); // Date header
			web_client_writef(node,"Server: %s\r\n",_libwebserver_version);
			web_client_writef(node,"Content-type: text/html\r\n\r\n<HTML><title>URL to large</title><body bgcolor=FFFFFF><font size=6>414 Requested url to large</font><BR><BR><HR><small><i>%s</i></small></body></html>\n\r",_libwebserver_version);                                    
			web_log("%s - - [%s] ^ URL to large\n",inet_ntoa(node->sa.sin_addr),mydate(time(NULL),"%d/%b/%Y:%H:%M:%S %z"));
			node->stat=5;
			IFDEBUG(fprintf(stderr,"client.c: (freeing getreq): ");fflush(stderr);); // REMOVE
			myfree(tmp1);
			return;
		};
		IFDEBUG(fprintf(stderr,"client.c: Matching %s to %s\n",matchbuf,tmp1));
		if(!fnmatch(matchbuf,tmp1,5)) {
			IFDEBUG(fprintf(stderr,"client.c: It matches\n"));	
			if((gettemp->flag & WS_LOCAL) == WS_LOCAL) {
				if(node->sa.sin_addr.s_addr!=0x0100007F) {
					web_client_writef(node,"HTTP/1.1 403 Forbidden\r\n");
					web_client_writef(node,"Date: %s\r\n",mydate(time(NULL),"%a, %d %b %Y %H:%M:%S %Z")); // Date header
					web_client_writef(node,"Server: %s\r\n",_libwebserver_version);
					web_client_writef(node,"Content-type: text/html\r\n\r\n<HTML><title>Forbidden</title><body bgcolor=FFFFFF><font size=6>403 Forbidden</font><BR><small>only local host accepted</small><BR><HR><small><i>%s</i></small></body></html>\n\r",_libwebserver_version);
					web_log("%s - - [%s] Connection rejected (only accepting 127.0.0.1)\n",inet_ntoa(node->sa.sin_addr),mydate(time(NULL),"%d/%b/%Y:%H:%M:%S %z"));
					node->stat=5;
					IFDEBUG(fprintf(stderr,"client.c: (freeing getreq): ");fflush(stderr);); // REMOVE
					myfree(tmp1);
					return;
				};
			};                        
			tmp=1; // Was found
		};
		IFDEBUG(fprintf(stderr,"client.c: (freeing getreq): ");fflush(stderr);); // REMOVE
	};	
	myfree(tmp1);
	if(!tmp) { // Nothing found
		web_client_writef(node,"HTTP/1.1 404 Not Found\r\n");
		web_client_writef(node,"Date: %s\r\n",mydate(time(NULL),"%a, %d %b %Y %H:%M:%S %Z")); // Date header
		web_client_writef(node,"Server: %s\r\n",_libwebserver_version);
		web_client_writef(node,"Content-type: text/html\r\n\r\n<HTML><title>not found</title><body bgcolor=FFFFFF><font size=6>404 NOT FOUND</font><BR><BR><HR><small><i>%s</i></small></body></html>\n\r",_libwebserver_version);                                    
		web_log("%s - - [%s] ^ (handler) Not found\n",inet_ntoa(node->sa.sin_addr),mydate(time(NULL),"%d/%b/%Y:%H:%M:%S %z"));
		node->stat=5;
	}else {
		fname=tmpfname();
		tfile=*stdout;
		if((nfile=freopen(fname,"wb+",stdout))!=NULL) {
			if(!add_outstream(node->outstream,fname,1)) {
				IFDEBUG(fprintf(stderr,"client.c:Couldn't add stream\n"));
			};
			// Setup Clientinfo before running function
			init_clientinfo();	 
			gettemp->func();
			free_clientinfo();        
			fflush(stdout);
			fclose(nfile); // it is a tempfile freopened
#ifdef WIN32			
			freopen("con","w",stdout);
#else
			freopen("/dev/tty","w",stdout);
#endif
			
		}else {
			fprintf(stderr,"Counldn't create temporary file %s\n",fname);
		};
		myfree(fname); // doesn't need anymore
		node->stat=4;   
		if(node->HTTPdirective==NULL) {
			web_client_writef(node,"HTTP/1.1 200 OK\r\n");
		} else {
			web_client_writef(node,"%s\r\n",node->HTTPdirective);
		};
		web_client_writef(node,"Date: %s\r\n",mydate(time(NULL),"%a, %d %b %Y %H:%M:%S %Z")); // Date header
		//Date: Mon, 14 Oct 2002 19:21:25 GMT
		web_client_writef(node,"Server: %s\r\n",_libwebserver_version);
	};

}
/*********************************************************************************************************/
/*
 * Process stream output
 */
void output_client(struct web_client *node) {
	struct outstream *tstream=node->outstream;
	char writetemp[WRITEMAX];
	if(tstream->next!=NULL) {
		if(tstream->next->fname!=NULL) {
			if(tstream->next->fstream==NULL) {
				if((tstream->next->fstream=fopen(tstream->next->fname,"rb"))==NULL) {
					IFDEBUG(fprintf(stderr,"client.c: Delecting stream\n"));
					delete_next_outstream(tstream);
					IFDEBUG(fprintf(stderr,"client.c: next=%p\n",tstream->next));
					node->outstream->next=tstream->next;
					//current_web_server->stat.bsent+=node->wrotesize; // stats
					node->wrotesize=0;
					return;
				};
			};
			if(tstream->next->wsize>0) {	// if last read ok
				if(tstream->next->rsize==0)fseek(tstream->next->fstream,0,SEEK_SET); // Exprimental
				tstream->next->rsize=fread(writetemp,1,WRITEMAX,tstream->next->fstream);
				IFDEBUG(fprintf(stderr,"client.c: Read %d bytes\n",tstream->next->rsize));
			} else {
				IFDEBUG(fprintf(stderr,"client.c: wsize is less or equal than 0 (didn't read)\n"););
			};
			if(tstream->next->rsize<1){
				fclose(tstream->next->fstream);
				IFDEBUG(fprintf(stderr,"client.c: Delecting stream\n"));
				delete_next_outstream(tstream);
				IFDEBUG(fprintf(stderr,"client.c: next=%p\n",tstream->next));
				node->outstream->next=tstream->next;
				//current_web_server->stat.bsent+=node->wrotesize; // stats
				node->wrotesize=0;
				return;
			};
			#ifdef HAVE_OPENSSL
				if(node->ssl!=NULL) {
					tstream->next->wsize=SSL_write(node->ssl,writetemp,tstream->next->rsize);
				} else {
					tstream->next->wsize=send(node->socket,writetemp,tstream->next->rsize,0);
				};
			#else		
				tstream->next->wsize=send(node->socket,writetemp,tstream->next->rsize,0);
			#endif
			IFDEBUG(fprintf(stderr,"client.c: wrote %d bytes\n",tstream->next->wsize));
			//write(2,writetemp,tstream->next->wsize); // to strerr 
			if(tstream->next->wsize>0) {
				IFDEBUG(fwrite(writetemp,tstream->next->wsize,1,stderr)); // to strerr 
				node->wrotesize+=tstream->next->wsize;	
				if(tstream->next->rsize!=tstream->next->wsize) {	
					fseek(tstream->next->fstream,node->wrotesize,SEEK_SET);
					IFDEBUG(fprintf(stderr,"client.c: Seeked to %d\n",node->wrotesize));
				};
			};
			if(tstream->next->wsize<=0 && errno!=11) {
				fclose(tstream->next->fstream);
				IFDEBUG(fprintf(stderr,"client.c: Delecting stream\n"));
				delete_next_outstream(tstream);
				IFDEBUG(fprintf(stderr,"client.c: next=%p\n",tstream->next));
				node->outstream->next=tstream->next;
				//current_web_server->stat.bsent+=node->wrotesize; // stats
				node->wrotesize=0;
				return;
			}else {
				if(tstream->next->wsize<0) {
					IFDEBUG(fprintf(stderr,"client.c: Broken pipe\n"););
				};
			};
		}else {;
			IFDEBUG(fprintf(stderr,"client.c: Delecting stream\n"));
			delete_next_outstream(tstream);
			IFDEBUG(fprintf(stderr,"client.c: next=%p\n",tstream->next));
			node->outstream->next=tstream->next;
			//current_web_server->stat.bsent+=node->wrotesize; // stats
			node->wrotesize=0;
			return;
		};
	}else { // End of streams
		current_web_client->stat=5; // done

	};
}

/*********************************************************************************************************/
/*
 * Set http directive
 */
void web_client_HTTPdirective(char *str) { 
	current_web_client->HTTPdirective=str;
}


/*********************************************************************************************************/
/*
 * GET request name
 */
char *web_client_getreq() {
	char *ret;
	char *tmp1=strstr(current_web_client->rbuf,"?");
	char *tmp2=strstr(current_web_client->rbuf," HTTP");
	char *tmp3=strstr(current_web_client->rbuf,"\r\n");
	int size;
	if(tmp1==NULL || tmp1>tmp2) {
		tmp1=tmp2;
	};
	if(tmp2>tmp3) {
		return NULL;
	};
	size=tmp1-current_web_client->rbuf;
	if(size<1) return NULL;
	IFDEBUG(fprintf(stderr,"client.c: (allocating getreq): ");fflush(stderr);); // REMOVE
	ret=mymalloc(size+1);
	if(ret==NULL) {
		fprintf(stderr,"client.c: Not enough mem to allocate getreq\n");
		return NULL;
	};
	memcpy(ret,current_web_client->rbuf,size);
	ret[size]=0;
	return ret;

};

/*********************************************************************************************************/
/*
 * GET request line
 */
char *web_client_getreqline() {
	char *ret;
	char *tmp1=strstr(current_web_client->rbuf,"\r\n");
	int size=tmp1-current_web_client->rbuf;
	if(size<1) return NULL;
	IFDEBUG(fprintf(stderr,"client.c: (allocating getreqline): ");fflush(stderr);); // REMOVE
	ret=mymalloc(size+1);
	if(ret==NULL) {
		fprintf(stderr,"client.c: Not enough mem to allocate getreqline\n");
		return NULL;
	};
	memcpy(ret,current_web_client->rbuf,size);
	ret[size]=0;
	return ret;
}


/*********************************************************************************************************/
/*
 * Add a FILE stream type to client output
 */
int web_client_addfile(char *in) {  
	int ret=add_outstream(current_web_client->outstream,in,0);
	
	char *fname=tmpfname();
	FILE *nfile;
	fflush(stdout);
	fclose(stdout); // oldstdout
	nfile=freopen(fname,"w+b",stdout);
	if(nfile!=NULL) { 
		if(!add_outstream(current_web_client->outstream,fname,1)) {
			IFDEBUG(fprintf(stderr,"client.c: Couldn't add stream\n"));
		};
	};                         	
	ClientInfo->outfd=fileno(stdout); // new output file descriptor
	return ret;
}


/*********************************************************************************************************/
/*
 * Output data as gif (with width w and height h)
 */
int web_client_outputgif(char *data,int w,int h) {
	int i;
	static int EGApalette[16][3] = {
			{0,0,0},       {0,0,128},     {0,128,0},     {0,128,128}, 
			{128,0,0},     {128,0,128},   {128,128,0},   {200,200,200},
			{100,100,100}, {100,100,255}, {100,255,100}, {100,255,255},
			{255,100,100}, {255,100,255}, {255,255,100}, {255,255,255} };
	char rm[256]={(char)0,(char)0xff}; //palette
	char gm[256]={(char)0,(char)0xff}; // palette
	char bm[256]={(char)0,(char)0xff}; // palette
	IFDEBUG(fprintf(stderr,"client.c: Creating EGA palette\n"));
	for (i=0; i<256; i++) {
		rm[i] = EGApalette[i&15][0];
		gm[i] = EGApalette[i&15][1];
		bm[i] = EGApalette[i&15][2];
	}
	IFDEBUG(fprintf(stderr,"client.c: Writting GIF\n"));
	i=WriteGIF(stdout,data,w,h,rm,gm,bm,256,0,"libwebserver export gif (xvgifwr.c)");
	IFDEBUG(fprintf(stderr,"client.c: GIF written\n"));
	return i;
};            


/*********************************************************************************************************/
/*
 * an util to write with format on client_nodes
 */
void web_client_writef(struct web_client *node,const char *fmt,...) {
	va_list args;
	char buf[WRITEMAX];
	va_start(args,fmt);
	vsnprintf(buf,512,fmt,args);
	va_end(args);
	IFDEBUG(fprintf(stderr,"client.c: wrote: %s\n",buf));
#ifdef HAVE_OPENSSL        
	if(node->ssl!=NULL) {
		SSL_write(node->ssl,buf,strlen(buf));
	} else {
		send(node->socket,buf,strlen(buf),0);   
	};
#else
	send(node->socket,buf,strlen(buf),0);
#endif
}
/*********************************************************************************************************/
/*
 * Set-cookie (new)
 */
void web_client_setcookie(char *n, char *v, char *t) {
	char *tmp1=t;
	int toffset=0;;
	while(*tmp1) {
		if(*tmp1=='S')toffset=1;             // seconds        
		if(*tmp1=='M')toffset=60;            // minutes
		if(*tmp1=='H')toffset=60*60;         // hours
		if(*tmp1=='d')toffset=60*60*24;      // days
		if(*tmp1=='m')toffset=60*60*24*30;   // Month
		if(*tmp1=='y')toffset=60*60*24*365;  // years
		tmp1++;
	};
	toffset+=timezone; // goes to GMT ,_timezone i guess is ansi
	printf("Set-Cookie: %s=%s; path=%s; expires=%s\r\n",n,v,ClientInfo->request,mydate(time(NULL)+(atoi(t)*toffset),"%a, %d-%b-%Y %H:%M:%S GMT"));
	
};
