/* Copyrights 2002 Luis Figueiredo (stdio@netc.pt) All rights reserved. 
 *
 * See the LICENSE file
 *
 * The origin of this software must not be misrepresented, either by
 * explicit claim or by omission.  Since few users ever read sources,
 * credits must appear in the documentation.
 *
 * date: Sat Mar 30 14:25:25 GMT 2002 
 *
 * -- memory functions
 */
#include "memory.h"

IFDEBUG(int _t=0;)


/*********************************************************************************************************/
/*
 * same as malloc with error reporting and libwebserver debug
 */
void * mymalloc(size_t s) {
	void *ret;
	ret=malloc(s);
	if(ret==NULL) {
		IFDEBUG(fprintf(stderr,"memory.c: malloc: %s (size %d)\n",strerror(errno),s));
		return NULL;
	};
	IFDEBUG(_t++;);
	IFDEBUG(fprintf(stderr,"memory.c: Allocated (%d) %d bytes to %p\n",_t,s,ret));
	return ret;
}

/*********************************************************************************************************/
/*
 * same as calloc with error reporting and libwebserver debug
 */
void * mycalloc(size_t nmemb,size_t s) {
	void *ret;
	ret=calloc(nmemb,s);
	if(ret==NULL) {
		IFDEBUG(fprintf(stderr,"memory.c: calloc %s\n",strerror(errno)));
		return NULL;
	};
	IFDEBUG(_t++;);
	IFDEBUG(fprintf(stderr,"memory.c: Allocated (%d) %d bytes to %p\n",_t,s*nmemb,ret));
	return ret;
}

/*********************************************************************************************************/
/* 
 * same as realloc with error reporting and libwebserver debug
 */
void * myrealloc(void *buf,size_t s) {
	void *ret;
	ret=realloc(buf,s);
#ifdef DEBUG
	if(buf==NULL) {
		_t++;
		IFDEBUG(fprintf(stderr,"memory.c: Allocated (%d) %d bytes to %p\n",_t,s,ret));
	};
#endif
	if(ret==NULL) {
		IFDEBUG(fprintf(stderr,"memory.c: realloc: %s\n",strerror(errno)));
		return NULL;
	};
	IFDEBUG(fprintf(stderr,"memory.c: Realloc buffer %p to %d\n",buf,s));
	return ret;
}


/*********************************************************************************************************/
/*
 * same as free with error report and libwebserver debug
 */ 
void myfree(void *ptr) {
	if(ptr!=NULL) {
		free(ptr);
		IFDEBUG(fprintf(stderr,"memory.c: Buffer (%d) %p freed\n",_t,ptr));
		IFDEBUG(_t--;);
	};
}


/*********************************************************************************************************/
/*
 *  Add a buffer to memrequest list
 */
void *add_buffer(struct memrequest *list,unsigned int size) {
	struct memrequest *tmem;
	if(size==0) {
		return NULL;
	};
	if(list!=NULL) {
		tmem=list;
	}else {
		return NULL;
	};
	while(tmem->next!=NULL)tmem=tmem->next;
	IFDEBUG(fprintf(stderr,"memory.c: (allocating mem buffer node): "));
	tmem->next=mymalloc(sizeof(struct memrequest));
	IFDEBUG(fprintf(stderr,"memory.c: (allocating mem buffer): "));
	tmem->next->ptr=mymalloc(size);
	tmem->next->next=NULL;
	return tmem->next->ptr;
}

/*********************************************************************************************************/
/*
 * Initialize memrequest list of buffers
 */
struct memrequest *init_buffer_list() {
	struct memrequest *newlist;
	IFDEBUG(fprintf(stderr,"memory.c: (allocating list): "));
	newlist=mymalloc(sizeof(struct memrequest));
	newlist->next=NULL;
	newlist->ptr=NULL;
	return newlist;
}

/*********************************************************************************************************/
/*
 * Delete memrequest buffer node (free)
 */
void delete_buffer(struct memrequest *mem) {
	IFDEBUG(fprintf(stderr,"memory.c: (freeing mem buffer): "));
	myfree(mem->ptr);
	IFDEBUG(fprintf(stderr,"memory.c: (freeing mem buffer node): "));
	myfree(mem);
}

/*********************************************************************************************************/
/*
 * Delete memrequest next buffer
 */
void delete_next_buffer(struct memrequest *mem) {
	struct memrequest *tmem;
	tmem=mem->next;
	mem->next=mem->next->next;
	delete_buffer(tmem);
}

/*********************************************************************************************************/
/*
 * Delete whole memrequest buffer list
 */
void delete_buffer_list(struct memrequest *list) {
	struct memrequest *tmem=list;
	if(tmem==NULL) return;

	while(tmem->next!=NULL) {
		delete_next_buffer(tmem);
	};
	delete_buffer(tmem);
}

