/* Copyrights 2002 Luis Figueiredo (stdio@netc.pt) All rights reserved. 
 *
 * See the LICENSE file
 *
 * The origin of this software must not be misrepresented, either by
 * explicit claim or by omission.  Since few users ever read sources,
 * credits must appear in the documentation.
 * 
 * date: Sat Mar 30 14:44:42 GMT 2002
 *
 * -- core server functions
 *
 */

#include "server.h"
#include "logo.h"

#ifdef WIN32
	#define _SERVER_VERSION "libwebserver 0.4.1 (win32)" // update allways when changing version (damn win)
#endif


#ifndef _SERVER_VERSION
	#define _SERVER_VERSION "libwebserver (unknow)"
#endif

char *_libwebserver_version= _SERVER_VERSION;

struct web_server *current_web_server;

/*********************************************************************************************************/
/*
 *	Define certificate file (open_ssl)
 */
void web_server_useSSLcert(struct web_server *server,const char *file) {
#ifdef HAVE_OPENSSL
	server->cert_file=mymalloc(strlen(file)+1);
	if(server->cert_file==NULL) {
		fprintf(stderr,"server.c: Not enough mem to allocate certificate file\n");
		return;
	};
	memcpy(server->cert_file,file,strlen(file));
	server->cert_file[strlen(file)]=0;
#endif
}

/*********************************************************************************************************/
/*
 *  Handler for libwebserver logotipe
 */
void _web_server_logo() {
	printf("Content-type: image/gif\r\n\r\n");
	fwrite((char *)_logo,sizeof(_logo),1,stdout);
}        


/*********************************************************************************************************/
/*
 * Add an handler to request data
 */
int web_server_addhandler(struct web_server *server,const char *mstr,void (*func)(),int flag) {
	_logfile=server->weblog;
	web_log("[%s] Adding handler %s\n",mydate(time(NULL),"%d/%b/%Y:%H:%M:%S %z"),mstr);
	if((flag & WS_LOCAL) == WS_LOCAL)web_log("                       +-- Only local connections\n");
	return add_handler((struct gethandler *)server->gethandler,mstr,func,flag);
}

/*********************************************************************************************************/
/*
 * This function initialize one web_server handler
 */
int web_server_init(struct web_server *server,int port,const char *logfile,int flags) {
#ifdef WIN32	
	WSADATA WSAinfo;
	WSAStartup(2,&WSAinfo); // Damn w32 sockets
#endif
	// Create a listen socket port 'port' and listen addr (0) (all interfaces)
	server->socket=listensocket((short)port,0);	
	if(server->socket==-1) {
		IFDEBUG(perror("listensocket"));
#ifdef WIN32		
		WSACleanup();
#endif
		return 0;
	};
	// Setup FILE structure of logfile
	server->weblog=NULL;
	if(logfile!=NULL) {
		IFDEBUG(fprintf(stderr,"server.c: Opening Log file %s\n",logfile));
		server->weblog=open_weblog(logfile); // Create File stream for log
		_logfile=server->weblog; // Set current log stream
		web_log("\n[%s] Server started at port %d (%s)\n",mydate(time(NULL),"%d/%b/%Y:%H:%M:%S %z"),port,_libwebserver_version);
	};
	// Setup Flags
	server->usessl=0;
	server->flags=flags;
	// openssl
#ifdef HAVE_OPENSSL	
	IFDEBUG(fprintf(stderr,"server.c: SSL stuff\n"));
	if((server->flags & WS_USESSL) == WS_USESSL) {
		web_log("[%s] (FLAG) Using SSL in connections\n",mydate(time(NULL),"%d/%b/%Y:%H:%M:%S %z"));	
		web_log("                       +-- %s certificate file\n",server->cert_file);
		SSL_load_error_strings();
		SSLeay_add_ssl_algorithms(); 	
		server->ctx=SSL_CTX_new (SSLv23_server_method());
		if (SSL_CTX_use_certificate_file(server->ctx, server->cert_file, SSL_FILETYPE_PEM) <= 0) {
			ERR_print_errors_fp(stderr);
			exit(3);
		}
		if (SSL_CTX_use_PrivateKey_file(server->ctx, server->cert_file, SSL_FILETYPE_PEM) <= 0) {
			ERR_print_errors_fp(stderr);
			exit(4);
		}                      
	 	if (SSL_CTX_check_private_key(server->ctx)<= 0)  	 {
			ERR_print_errors_fp(stderr);
			exit(4);
		};
		server->usessl=1;
	};
#endif

	server->client=init_client_list();										// Initializate client list
	server->gethandler=init_handler_list();									// Initializate handlers list
	web_server_addhandler(server,"* /libwebserver.gif",_web_server_logo,0);	// Add logo default handler
#ifndef WIN32	
	signal(SIGPIPE,SIG_IGN);
#endif
	return 1;
}                            


/*********************************************************************************************************/
/*
 * Core function
 */
int web_server_run(struct web_server *server) {
	struct web_client *client;
	char *tmp;
	_logfile=server->weblog;
	current_web_server=server;
	if(newdata(server->socket)) {
		IFDEBUG(fprintf(stderr,"server.c: Client connected\n"));
		IFDEBUG(fprintf(stderr,"server.c:(allocating client): ");fflush(stderr);); // REMOVE
		client=mymalloc(sizeof(struct web_client));
		if(client==NULL) {
			fprintf(stderr,"server.c: Not enough mem to allocate client\n");
			return 0;
		};
		client->salen=sizeof(client->sa);
		client->socket=accept(server->socket,(struct sockaddr *)&client->sa,&client->salen);
		current_web_server->stat.connection+=1;
#ifdef HAVE_OPENSSL
		if((server->flags & WS_USESSL) == WS_USESSL) {
			client->ssl = SSL_new(server->ctx);
			SSL_set_fd(client->ssl,client->socket);
			SSL_accept(client->ssl);
			//client->cert = SSL_get_peer_certificate (client->ssl);
		} else {
			client->ssl=NULL;
		};
#endif
		if(client->socket==-1) {
			IFDEBUG(perror("accept"));
			IFDEBUG(fprintf(stderr,"server.c: %d client->sa=%p client->salen=%d\n",server->socket,&client->sa,client->salen));
			IFDEBUG(fprintf(stderr,"server.c:(freeing client): ");fflush(stderr);); // REMOVE
			myfree(client);
			return 0;
		};
		if(!add_client(server->client,client)) {
			fprintf(stderr,"server.c: Error adding client\n");
			return 0;
		};
	};
	client=server->client;
	if(!client->next) { // think of Rocco Carbone (rocco@tecsiel.it)
		return 2; // i don't need to process the list (nothing next) returns 2 if there is no client to process
	};
	while(client->next!=NULL) { // Process the client and swap to next;
		current_web_client=client->next;
		switch(client->next->stat) {
			case 1: {
				read_client(current_web_client);
			};break;
            case 2: {
				tmp=web_client_getreqline();
				web_log("%s - - [%s] \"%s\"\n",inet_ntoa(current_web_client->sa.sin_addr),mydate(time(NULL),"%d/%b/%Y:%H:%M:%S %z"),tmp); 
				IFDEBUG(fprintf(stderr,"server.c:(freeing getreqline): ");fflush(stderr);); // REMOVE
				myfree(tmp);
				process_client(current_web_client,server->gethandler);
			};break;
			case 4: {
				output_client(current_web_client);	
			};break;
			case 5: {
				delete_next_client(client); 
				continue;
			};break;
		};
		client=client->next;
	};                                       	
	return 1;
}


